@@ -5,9 +5,11 @@ use crate::args::FontsCommand;
 
 /// Execute a font listing command.
 pub fn fonts(command: &FontsCommand) {
-    let fonts = Fonts::searcher()
-        .include_system_fonts(!command.font_args.ignore_system_fonts)
-        .search_with(&command.font_args.font_paths);
+    let mut fonts = Fonts::searcher();
+    fonts.include_system_fonts(!command.font.ignore_system_fonts);
+    #[cfg(feature = "embed-fonts")]
+    fonts.include_embedded_fonts(!command.font.ignore_embedded_fonts);
+    let fonts = fonts.search_with(&command.font.font_paths);
 
     for (name, infos) in fonts.book.families() {
         println!("{name}");
