@@ -2,13 +2,16 @@ use std::fmt::{self, Debug, Formatter};
 use std::ops::{Deref, DerefMut};
 
 use typst_library::engine::Engine;
-use typst_library::introspection::{SplitLocator, Tag};
+use typst_library::foundations::Resolve;
+use typst_library::introspection::{SplitLocator, Tag, TagFlags};
 use typst_library::layout::{Abs, Dir, Em, Fr, Frame, FrameItem, Point};
 use typst_library::model::ParLineMarker;
-use typst_library::text::{Lang, TextElem};
+use typst_library::text::{Lang, TextElem, variant};
 use typst_utils::Numeric;
 
 use super::*;
+use crate::inline::linebreak::Trim;
+use crate::inline::shaping::Adjustability;
 use crate::modifiers::layout_and_modify;
 
 const SHY: char = '\u{ad}';
@@ -60,8 +63,7 @@ impl Line<'_> {
         // CJK character at line end should not be adjusted.
         if self
             .items
-            .last()
-            .and_then(Item::text)
+            .trailing_text()
             .map(|s| s.cjk_justifiable_at_last())
             .unwrap_or(false)
         {
@@ -127,7 +129,7 @@ pub fn line<'a>(
     p: &'a Preparation,
     range: Range,
     breakpoint: Breakpoint,
-    pred: Option<&Line>,
+    pred: Option<&Line<'a>>,
 ) -> Line<'a> {
     // The line's full text.
     let full = &p.text[range.clone()];
@@ -148,27 +150,43 @@ pub fn line<'a>(
     };
 
     // Trim the line at the end, if necessary for this breakpoint.
-    let trim = range.start + breakpoint.trim(full).len();
+    let trim = breakpoint.trim(range.start, full);
+    let trimmed_range = range.start..trim.layout;
 
     // Collect the items for the line.
-    let mut items = collect_items(engine, p, range, trim);
+    let mut items = Items::new();
 
     // Add a hyphen at the line start, if a previous dash should be repeated.
-    if pred.map_or(false, |pred| should_repeat_hyphen(pred, full)) {
-        if let Some(shaped) = items.first_text_mut() {
-            shaped.prepend_hyphen(engine, p.config.fallback);
-        }
+    if let Some(pred) = pred
+        && pred.dash == Some(Dash::Hard)
+        && let Some(base) = pred.items.trailing_text()
+        && should_repeat_hyphen(base.lang, full)
+        && let Some(hyphen) =
+            ShapedText::hyphen(engine, p.config.fallback, base, trim.shaping, false)
+    {
+        items.push(Item::Text(hyphen), LogicalIndex::START_HYPHEN);
     }
 
+    collect_items(&mut items, engine, p, range, &trim);
+
     // Add a hyphen at the line end, if we ended on a soft hyphen.
-    if dash == Some(Dash::Soft) {
-        if let Some(shaped) = items.last_text_mut() {
-            shaped.push_hyphen(engine, p.config.fallback);
-        }
+    if dash == Some(Dash::Soft)
+        && let Some(base) = items.trailing_text()
+        && let Some(hyphen) =
+            ShapedText::hyphen(engine, p.config.fallback, base, trim.shaping, true)
+    {
+        items.push(Item::Text(hyphen), LogicalIndex::END_HYPHEN);
     }
 
+    // Ensure that there is no weak spacing at the start and end of the line.
+    trim_weak_spacing(&mut items);
+
     // Deal with CJ characters at line boundaries.
-    adjust_cj_at_line_boundaries(p, full, &mut items);
+    // Use the trimmed range for robust boundary checks.
+    adjust_cj_at_line_boundaries(p, trimmed_range, &mut items);
+
+    // Deal with stretchability of glyphs at the end of the line.
+    adjust_glyph_stretch_at_line_end(p, &mut items);
 
     // Compute the line's width.
     let width = items.iter().map(Item::natural_width).sum();
@@ -185,23 +203,33 @@ pub fn line<'a>(
 /// We do not factor the `trim` directly into the `range` because we still want
 /// to keep non-text items after the trim (e.g. tags).
 fn collect_items<'a>(
+    items: &mut Items<'a>,
     engine: &Engine,
     p: &'a Preparation,
     range: Range,
-    trim: usize,
-) -> Items<'a> {
-    let mut items = Items::new();
+    trim: &Trim,
+) {
     let mut fallback = None;
 
     // Collect the items for each consecutively ordered run.
     reorder(p, range.clone(), |subrange, rtl| {
         let from = items.len();
-        collect_range(engine, p, subrange, trim, &mut items, &mut fallback);
+        collect_range(engine, p, subrange, trim, items, &mut fallback);
         if rtl {
             items.reorder(from);
         }
     });
 
+    // Add fallback text to expand the line height, if necessary.
+    if !items.iter().any(|item| matches!(item, Item::Text(_)))
+        && let Some(fallback) = fallback
+    {
+        items.push(fallback, LogicalIndex::FALLBACK_TEXT);
+    }
+}
+
+/// Trims weak spacing from the start and end of the line.
+fn trim_weak_spacing(items: &mut Items) {
     // Trim weak spacing at the start of the line.
     let prefix = items
         .iter()
@@ -212,18 +240,9 @@ fn collect_items<'a>(
     }
 
     // Trim weak spacing at the end of the line.
-    while matches!(items.last(), Some(Item::Absolute(_, true))) {
+    while matches!(items.iter().next_back(), Some(Item::Absolute(_, true))) {
         items.pop();
     }
-
-    // Add fallback text to expand the line height, if necessary.
-    if !items.iter().any(|item| matches!(item, Item::Text(_))) {
-        if let Some(fallback) = fallback {
-            items.push(fallback);
-        }
-    }
-
-    items
 }
 
 /// Calls `f` for the BiDi-reordered ranges of a line.
@@ -266,21 +285,23 @@ fn collect_range<'a>(
     engine: &Engine,
     p: &'a Preparation,
     range: Range,
-    trim: usize,
+    trim: &Trim,
     items: &mut Items<'a>,
     fallback: &mut Option<ItemEntry<'a>>,
 ) {
-    for (subrange, item) in p.slice(range.clone()) {
+    for (i, (subrange, item)) in p.slice(range.clone()) {
+        let idx = LogicalIndex::from_item_index(i);
+
         // All non-text items are just kept, they can't be split.
         let Item::Text(shaped) = item else {
-            items.push(item);
+            items.push(item, idx);
             continue;
         };
 
         // The intersection range of the item, the subrange, and the line's
         // trimming.
-        let sliced =
-            range.start.max(subrange.start)..range.end.min(subrange.end).min(trim);
+        let sliced = range.start.max(subrange.start)
+            ..range.end.min(subrange.end).min(trim.shaping);
 
         // Whether the item is split by the line.
         let split = subrange.start < sliced.start || sliced.end < subrange.end;
@@ -290,14 +311,25 @@ fn collect_range<'a>(
             // we can use to force a non-zero line-height when the line doesn't
             // contain any other text.
             *fallback = Some(ItemEntry::from(Item::Text(shaped.empty())));
-        } else if split {
+            continue;
+        }
+
+        let mut item: ItemEntry = if split {
             // When the item is split in half, reshape it.
             let reshaped = shaped.reshape(engine, sliced);
-            items.push(Item::Text(reshaped));
+            Item::Text(reshaped).into()
         } else {
             // When the item is fully contained, just keep it.
-            items.push(item);
+            item.into()
+        };
+
+        // Trim end-of-line whitespace glyphs.
+        if trim.layout < range.end {
+            let shaped = item.text_mut().unwrap();
+            shaped.glyphs.trim(|glyph| trim.layout < glyph.range.end);
         }
+
+        items.push(item, idx);
     }
 }
 
@@ -305,7 +337,11 @@ fn collect_range<'a>(
 ///
 /// See Requirements for Chinese Text Layout, Section 3.1.6.3 Compression of
 /// punctuation marks at line start or line end.
-fn adjust_cj_at_line_boundaries(p: &Preparation, text: &str, items: &mut Items) {
+///
+/// The `range` should only contain regular texts, with linebreaks trimmed.
+fn adjust_cj_at_line_boundaries(p: &Preparation, range: Range, items: &mut Items) {
+    let text = &p.text[range];
+
     if text.starts_with(BEGIN_PUNCT_PAT)
         || (p.config.cjk_latin_spacing && text.starts_with(is_of_cj_script))
     {
@@ -319,9 +355,25 @@ fn adjust_cj_at_line_boundaries(p: &Preparation, text: &str, items: &mut Items)
     }
 }
 
+/// Remove stretchability from the last glyph in the line to avoid trailing
+/// space after a glyph due to glyph-level justification.
+fn adjust_glyph_stretch_at_line_end(p: &Preparation, items: &mut Items) {
+    let glyph_limits = &p.config.justification_limits.tracking();
+    if glyph_limits.min.is_zero() && glyph_limits.max.is_zero() {
+        return;
+    }
+
+    // This is not perfect (ignores clusters and is not particularly fast), but
+    // it is good enough for now. The adjustability handling in general needs a
+    // cleanup.
+    let Some(shaped) = items.trailing_text_mut() else { return };
+    let Some(glyph) = shaped.glyphs.to_mut().last_mut() else { return };
+    glyph.adjustability = Adjustability::default();
+}
+
 /// Add spacing around punctuation marks for CJ glyphs at the line start.
 fn adjust_cj_at_line_start(p: &Preparation, items: &mut Items) {
-    let Some(shaped) = items.first_text_mut() else { return };
+    let Some(shaped) = items.leading_text_mut() else { return };
     let Some(glyph) = shaped.glyphs.first() else { return };
 
     if glyph.is_cjk_right_aligned_punctuation() {
@@ -330,7 +382,6 @@ fn adjust_cj_at_line_start(p: &Preparation, items: &mut Items) {
         let glyph = shaped.glyphs.to_mut().first_mut().unwrap();
         let shrink = glyph.shrinkability().0;
         glyph.shrink_left(shrink);
-        shaped.width -= shrink.at(shaped.size);
     } else if p.config.cjk_latin_spacing
         && glyph.is_cj_script()
         && glyph.x_offset > Em::zero()
@@ -342,13 +393,12 @@ fn adjust_cj_at_line_start(p: &Preparation, items: &mut Items) {
         glyph.x_advance -= shrink;
         glyph.x_offset = Em::zero();
         glyph.adjustability.shrinkability.0 = Em::zero();
-        shaped.width -= shrink.at(shaped.size);
     }
 }
 
 /// Add spacing around punctuation marks for CJ glyphs at the line end.
 fn adjust_cj_at_line_end(p: &Preparation, items: &mut Items) {
-    let Some(shaped) = items.last_text_mut() else { return };
+    let Some(shaped) = items.trailing_text_mut() else { return };
     let Some(glyph) = shaped.glyphs.last() else { return };
 
     // Deal with CJK punctuation at line ends.
@@ -360,7 +410,6 @@ fn adjust_cj_at_line_end(p: &Preparation, items: &mut Items) {
         let shrink = glyph.shrinkability().1;
         let punct = shaped.glyphs.to_mut().last_mut().unwrap();
         punct.shrink_right(shrink);
-        shaped.width -= shrink.at(shaped.size);
     } else if p.config.cjk_latin_spacing
         && glyph.is_cj_script()
         && (glyph.x_advance - glyph.x_offset) > Em::one()
@@ -371,23 +420,13 @@ fn adjust_cj_at_line_end(p: &Preparation, items: &mut Items) {
         let glyph = shaped.glyphs.to_mut().last_mut().unwrap();
         glyph.x_advance -= shrink;
         glyph.adjustability.shrinkability.1 = Em::zero();
-        shaped.width -= shrink.at(shaped.size);
     }
 }
 
-/// Whether a hyphen should be inserted at the start of the next line.
-fn should_repeat_hyphen(pred_line: &Line, text: &str) -> bool {
-    // If the predecessor line does not end with a `Dash::Hard`, we shall
-    // not place a hyphen at the start of the next line.
-    if pred_line.dash != Some(Dash::Hard) {
-        return false;
-    }
-
-    // The hyphen should repeat only in the languages that require that feature.
-    // For more information see the discussion at https://github.com/typst/typst/issues/3235
-    let Some(Item::Text(shaped)) = pred_line.items.last() else { return false };
-
-    match shaped.lang {
+/// Whether a hyphen should be repeated at the start of the line in the given
+/// language, when the following text is the given one.
+fn should_repeat_hyphen(lang: Lang, following_text: &str) -> bool {
+    match lang {
         // - Lower Sorbian: see https://dolnoserbski.de/ortografija/psawidla/K3
         // - Czech: see https://prirucka.ujc.cas.cz/?id=164
         // - Croatian: see http://pravopis.hr/pravilo/spojnica/68/
@@ -406,15 +445,37 @@ fn should_repeat_hyphen(pred_line: &Line, text: &str) -> bool {
         //
         // See § 4.1.1.1.2.e on the "Ortografía de la lengua española"
         // https://www.rae.es/ortografía/como-signo-de-división-de-palabras-a-final-de-línea
-        Lang::SPANISH => text.chars().next().map_or(false, |c| !c.is_uppercase()),
+        Lang::SPANISH => following_text.chars().next().is_some_and(|c| !c.is_uppercase()),
 
         _ => false,
     }
 }
 
-/// Apply the current baseline shift to a frame.
-pub fn apply_baseline_shift(frame: &mut Frame, styles: StyleChain) {
-    frame.translate(Point::with_y(TextElem::baseline_in(styles)));
+/// Apply the current baseline shift and italic compensation to a frame.
+pub fn apply_shift<'a>(
+    world: &Tracked<'a, dyn World + 'a>,
+    frame: &mut Frame,
+    styles: StyleChain,
+) {
+    let mut baseline = styles.resolve(TextElem::baseline);
+    let mut compensation = Abs::zero();
+    if let Some(scripts) = styles.get_ref(TextElem::shift_settings) {
+        let font_metrics = styles
+            .get_ref(TextElem::font)
+            .into_iter()
+            .find_map(|family| {
+                world
+                    .book()
+                    .select(family.as_str(), variant(styles))
+                    .and_then(|id| world.font(id))
+            })
+            .map_or(*scripts.kind.default_metrics(), |f| {
+                *scripts.kind.read_metrics(f.metrics())
+            });
+        baseline -= scripts.shift.unwrap_or(font_metrics.vertical_offset).resolve(styles);
+        compensation += font_metrics.horizontal_offset.resolve(styles);
+    }
+    frame.translate(Point::new(compensation, baseline));
 }
 
 /// Commit to a line and build its frame.
@@ -438,30 +499,26 @@ pub fn commit(
     }
 
     // Handle hanging punctuation to the left.
-    if let Some(Item::Text(text)) = line.items.first() {
-        if let Some(glyph) = text.glyphs.first() {
-            if !text.dir.is_positive()
-                && TextElem::overhang_in(text.styles)
-                && (line.items.len() > 1 || text.glyphs.len() > 1)
-            {
-                let amount = overhang(glyph.c) * glyph.x_advance.at(text.size);
-                offset -= amount;
-                remaining += amount;
-            }
-        }
+    if let Some(text) = line.items.leading_text()
+        && let Some(glyph) = text.glyphs.first()
+        && !text.dir.is_positive()
+        && text.styles.get(TextElem::overhang)
+        && (line.items.len() > 1 || text.glyphs.len() > 1)
+    {
+        let amount = overhang(glyph.c) * glyph.x_advance.at(glyph.size);
+        offset -= amount;
+        remaining += amount;
     }
 
     // Handle hanging punctuation to the right.
-    if let Some(Item::Text(text)) = line.items.last() {
-        if let Some(glyph) = text.glyphs.last() {
-            if text.dir.is_positive()
-                && TextElem::overhang_in(text.styles)
-                && (line.items.len() > 1 || text.glyphs.len() > 1)
-            {
-                let amount = overhang(glyph.c) * glyph.x_advance.at(text.size);
-                remaining += amount;
-            }
-        }
+    if let Some(text) = line.items.trailing_text()
+        && let Some(glyph) = text.glyphs.last()
+        && text.dir.is_positive()
+        && text.styles.get(TextElem::overhang)
+        && (line.items.len() > 1 || text.glyphs.len() > 1)
+    {
+        let amount = overhang(glyph.c) * glyph.x_advance.at(glyph.size);
+        remaining += amount;
     }
 
     // Determine how much additional space is needed. The justification_ratio is
@@ -499,16 +556,16 @@ pub fn commit(
 
     // Build the frames and determine the height and baseline.
     let mut frames = vec![];
-    for item in line.items.iter() {
-        let mut push = |offset: &mut Abs, frame: Frame| {
+    for &(idx, ref item) in line.items.indexed_iter() {
+        let mut push = |offset: &mut Abs, frame: Frame, idx: LogicalIndex| {
             let width = frame.width();
             top.set_max(frame.baseline());
             bottom.set_max(frame.size().y - frame.baseline());
-            frames.push((*offset, frame));
+            frames.push((*offset, frame, idx));
             *offset += width;
         };
 
-        match item {
+        match &**item {
             Item::Absolute(v, _) => {
                 offset += *v;
             }
@@ -519,8 +576,8 @@ pub fn commit(
                     let mut frame = layout_and_modify(*styles, |styles| {
                         layout_box(elem, engine, loc.relayout(), styles, region)
                     })?;
-                    apply_baseline_shift(&mut frame, *styles);
-                    push(&mut offset, frame);
+                    apply_shift(&engine.world, &mut frame, *styles);
+                    push(&mut offset, frame, idx);
                 } else {
                     offset += amount;
                 }
@@ -532,15 +589,15 @@ pub fn commit(
                     justification_ratio,
                     extra_justification,
                 );
-                push(&mut offset, frame);
+                push(&mut offset, frame, idx);
             }
             Item::Frame(frame) => {
-                push(&mut offset, frame.clone());
+                push(&mut offset, frame.clone(), idx);
             }
             Item::Tag(tag) => {
                 let mut frame = Frame::soft(Size::zero());
                 frame.push(Point::zero(), FrameItem::Tag((*tag).clone()));
-                frames.push((offset, frame));
+                frames.push((offset, frame, idx));
             }
             Item::Skip(_) => {}
         }
@@ -559,8 +616,13 @@ pub fn commit(
         add_par_line_marker(&mut output, marker, engine, locator, top);
     }
 
+    // Ensure that the final frame's items are in logical order rather than in
+    // visual order. This is important because it affects the order of elements
+    // during introspection and thus things like counters.
+    frames.sort_unstable_by_key(|(_, _, idx)| *idx);
+
     // Construct the line's frame.
-    for (offset, frame) in frames {
+    for (offset, frame, _) in frames {
         let x = offset + p.config.align.position(remaining);
         let y = top - frame.baseline();
         output.push_frame(Point::new(x, y), frame);
@@ -601,8 +663,9 @@ fn add_par_line_marker(
     // line's general baseline. However, the line number will still need to
     // manually adjust its own 'y' position based on its own baseline.
     let pos = Point::with_y(top);
-    output.push(pos, FrameItem::Tag(Tag::Start(marker.pack())));
-    output.push(pos, FrameItem::Tag(Tag::End(loc, key)));
+    let flags = TagFlags { introspectable: false, tagged: false };
+    output.push(pos, FrameItem::Tag(Tag::Start(marker.pack(), flags)));
+    output.push(pos, FrameItem::Tag(Tag::End(loc, key, flags)));
 }
 
 /// How much a character should hang into the end margin.
@@ -613,7 +676,7 @@ fn overhang(c: char) -> f64 {
     match c {
         // Dashes.
         '–' | '—' => 0.2,
-        '-' => 0.55,
+        '-' | '\u{ad}' => 0.55,
 
         // Punctuation.
         '.' | ',' => 0.8,
@@ -627,7 +690,7 @@ fn overhang(c: char) -> f64 {
 }
 
 /// A collection of owned or borrowed inline items.
-pub struct Items<'a>(Vec<ItemEntry<'a>>);
+pub struct Items<'a>(Vec<(LogicalIndex, ItemEntry<'a>)>);
 
 impl<'a> Items<'a> {
     /// Create empty items.
@@ -636,33 +699,44 @@ impl<'a> Items<'a> {
     }
 
     /// Push a new item.
-    pub fn push(&mut self, entry: impl Into<ItemEntry<'a>>) {
-        self.0.push(entry.into());
+    pub fn push(&mut self, entry: impl Into<ItemEntry<'a>>, idx: LogicalIndex) {
+        self.0.push((idx, entry.into()));
     }
 
-    /// Iterate over the items
-    pub fn iter(&self) -> impl Iterator<Item = &Item<'a>> {
-        self.0.iter().map(|item| &**item)
+    /// Iterate over the items.
+    pub fn iter(&self) -> impl DoubleEndedIterator<Item = &Item<'a>> {
+        self.0.iter().map(|(_, item)| &**item)
     }
 
-    /// Access the first item.
-    pub fn first(&self) -> Option<&Item<'a>> {
-        self.0.first().map(|item| &**item)
+    /// Iterate over the items with the indices that define their logical order.
+    /// See the docs above `logical_item_idx` for more details.
+    ///
+    /// Note that this is different from `.iter().enumerate()` which would
+    /// provide the indices in visual order!
+    pub fn indexed_iter(
+        &self,
+    ) -> impl DoubleEndedIterator<Item = &(LogicalIndex, ItemEntry<'a>)> {
+        self.0.iter()
     }
 
-    /// Access the last item.
-    pub fn last(&self) -> Option<&Item<'a>> {
-        self.0.last().map(|item| &**item)
+    /// Access the first item (skipping tags), if it is text.
+    pub fn leading_text(&self) -> Option<&ShapedText<'a>> {
+        self.0.iter().find(|(_, item)| !item.is_tag())?.1.text()
     }
 
-    /// Access the first item mutably, if it is text.
-    pub fn first_text_mut(&mut self) -> Option<&mut ShapedText<'a>> {
-        self.0.first_mut()?.text_mut()
+    /// Access the first item (skipping tags) mutably, if it is text.
+    pub fn leading_text_mut(&mut self) -> Option<&mut ShapedText<'a>> {
+        self.0.iter_mut().find(|(_, item)| !item.is_tag())?.1.text_mut()
     }
 
-    /// Access the last item mutably, if it is text.
-    pub fn last_text_mut(&mut self) -> Option<&mut ShapedText<'a>> {
-        self.0.last_mut()?.text_mut()
+    /// Access the last item (skipping tags), if it is text.
+    pub fn trailing_text(&self) -> Option<&ShapedText<'a>> {
+        self.0.iter().rev().find(|(_, item)| !item.is_tag())?.1.text()
+    }
+
+    /// Access the last item (skipping tags) mutably, if it is text.
+    pub fn trailing_text_mut(&mut self) -> Option<&mut ShapedText<'a>> {
+        self.0.iter_mut().rev().find(|(_, item)| !item.is_tag())?.1.text_mut()
     }
 
     /// Reorder the items starting at the given index to RTL.
@@ -671,14 +745,8 @@ impl<'a> Items<'a> {
     }
 }
 
-impl<'a> FromIterator<ItemEntry<'a>> for Items<'a> {
-    fn from_iter<I: IntoIterator<Item = ItemEntry<'a>>>(iter: I) -> Self {
-        Self(iter.into_iter().collect())
-    }
-}
-
 impl<'a> Deref for Items<'a> {
-    type Target = Vec<ItemEntry<'a>>;
+    type Target = Vec<(LogicalIndex, ItemEntry<'a>)>;
 
     fn deref(&self) -> &Self::Target {
         &self.0
@@ -697,7 +765,39 @@ impl Debug for Items<'_> {
     }
 }
 
+/// We use indices to remember the logical (as opposed to visual) order of
+/// items. During line building, the items are stored in visual (BiDi-reordered)
+/// order. When committing to a line and building its frame, we sort by logical
+/// index.
+///
+/// - Special layout-generated items have custom indices that ensure correct
+///   ordering w.r.t. to each other and normal elements, listed below.
+/// - Normal items have their position in `p.items` plus the number of special
+///   reserved prefix indices.
+///
+/// Logical indices must be unique within a line because we use an unstable
+/// sort.
+#[derive(Debug, Copy, Clone, Eq, PartialEq, Ord, PartialOrd)]
+pub struct LogicalIndex(usize);
+
+impl LogicalIndex {
+    const START_HYPHEN: Self = Self(0);
+    const FALLBACK_TEXT: Self = Self(usize::MAX - 1);
+    const END_HYPHEN: Self = Self(usize::MAX);
+
+    /// Create a logical index from the index of an item in the [`p.items`](Preparation::items).
+    const fn from_item_index(i: usize) -> Self {
+        // This won't overflow because the `idx` comes from a vector which is
+        // limited to `isize::MAX` elements.
+        Self(i + 1)
+    }
+}
+
 /// A reference to or a boxed item.
+///
+/// This is conceptually similar to a [`Cow<'a, Item<'a>>`][std::borrow::Cow],
+/// but we box owned items since an [`Item`] is much bigger than
+/// a box.
 pub enum ItemEntry<'a> {
     Ref(&'a Item<'a>),
     Box(Box<Item<'a>>),
