@@ -4,18 +4,16 @@ use std::num::{NonZeroI64, NonZeroUsize};
 use std::ops::{Add, AddAssign};
 
 use comemo::Tracked;
-use ecow::{eco_format, EcoString, EcoVec};
+use ecow::{EcoString, EcoVec, eco_format};
 use serde::{Deserialize, Serialize};
 use smallvec::SmallVec;
 use typst_syntax::{Span, Spanned};
 
-use crate::diag::{
-    bail, At, DeprecationSink, HintedStrResult, SourceDiagnostic, SourceResult, StrResult,
-};
+use crate::diag::{At, HintedStrResult, SourceDiagnostic, SourceResult, StrResult, bail};
 use crate::engine::Engine;
 use crate::foundations::{
-    cast, func, ops, repr, scope, ty, Args, Bytes, CastInfo, Context, Dict, FromValue,
-    Func, IntoValue, Reflect, Repr, Str, Value, Version,
+    Args, Bytes, CastInfo, Context, Dict, FromValue, Func, IntoValue, Reflect, Repr, Str,
+    Value, Version, cast, func, ops, repr, scope, ty,
 };
 
 /// Create a new [`Array`] from values.
@@ -96,7 +94,7 @@ impl Array {
     }
 
     /// Iterate over references to the contained values.
-    pub fn iter(&self) -> std::slice::Iter<Value> {
+    pub fn iter(&self) -> std::slice::Iter<'_, Value> {
         self.0.iter()
     }
 
@@ -145,11 +143,6 @@ impl Array {
 
         Ok(self.iter().cloned().cycle().take(count).collect())
     }
-
-    /// The internal implementation of [`Array::contains`].
-    pub fn contains_impl(&self, value: &Value, sink: &mut dyn DeprecationSink) -> bool {
-        self.0.iter().any(|v| ops::equal(v, value, sink))
-    }
 }
 
 #[scope]
@@ -179,17 +172,29 @@ impl Array {
     }
 
     /// Returns the first item in the array. May be used on the left-hand side
-    /// of an assignment. Fails with an error if the array is empty.
+    /// an assignment. Returns the default value if the array is empty
+    /// or fails with an error is no default value was specified.
     #[func]
-    pub fn first(&self) -> StrResult<Value> {
-        self.0.first().cloned().ok_or_else(array_is_empty)
+    pub fn first(
+        &self,
+        /// A default value to return if the array is empty.
+        #[named]
+        default: Option<Value>,
+    ) -> StrResult<Value> {
+        self.0.first().cloned().or(default).ok_or_else(array_is_empty)
     }
 
     /// Returns the last item in the array. May be used on the left-hand side of
-    /// an assignment. Fails with an error if the array is empty.
+    /// an assignment. Returns the default value if the array is empty
+    /// or fails with an error is no default value was specified.
     #[func]
-    pub fn last(&self) -> StrResult<Value> {
-        self.0.last().cloned().ok_or_else(array_is_empty)
+    pub fn last(
+        &self,
+        /// A default value to return if the array is empty.
+        #[named]
+        default: Option<Value>,
+    ) -> StrResult<Value> {
+        self.0.last().cloned().or(default).ok_or_else(array_is_empty)
     }
 
     /// Returns the item at the specified index in the array. May be used on the
@@ -281,11 +286,8 @@ impl Array {
         #[named]
         count: Option<i64>,
     ) -> StrResult<Array> {
-        let mut end = end;
-        if end.is_none() {
-            end = count.map(|c: i64| start + c);
-        }
         let start = self.locate(start, true)?;
+        let end = end.or(count.map(|c| start as i64 + c));
         let end = self.locate(end.unwrap_or(self.len() as i64), true)?.max(start);
         Ok(self.0[start..end].into())
     }
@@ -297,12 +299,10 @@ impl Array {
     #[func]
     pub fn contains(
         &self,
-        engine: &mut Engine,
-        span: Span,
         /// The value to search for.
         value: Value,
     ) -> bool {
-        self.contains_impl(&value, &mut (engine, span))
+        self.0.contains(&value)
     }
 
     /// Searches for an item for which the given function returns `{true}` and
@@ -443,6 +443,14 @@ impl Array {
     /// The returned array consists of `(index, value)` pairs in the form of
     /// length-2 arrays. These can be [destructured]($scripting/#bindings) with
     /// a let binding or for loop.
+    ///
+    /// ```example
+    /// #for (i, value) in ("A", "B", "C").enumerate() {
+    ///   [#i: #value \ ]
+    /// }
+    ///
+    /// #("A", "B", "C").enumerate(start: 1)
+    /// ```
     #[func]
     pub fn enumerate(
         self,
@@ -563,6 +571,11 @@ impl Array {
     }
 
     /// Folds all items into a single value using an accumulator function.
+    ///
+    /// ```example
+    /// #let array = (1, 2, 3, 4)
+    /// #array.fold(0, (acc, x) => acc + x)
+    /// ```
     #[func]
     pub fn fold(
         self,
@@ -585,8 +598,6 @@ impl Array {
     #[func]
     pub fn sum(
         self,
-        engine: &mut Engine,
-        span: Span,
         /// What to return if the array is empty. Must be set if the array can
         /// be empty.
         #[named]
@@ -598,12 +609,12 @@ impl Array {
             .or(default)
             .ok_or("cannot calculate sum of empty array with no default")?;
         for item in iter {
-            acc = ops::add(acc, item, &mut (&mut *engine, span))?;
+            acc = ops::add(acc, item)?;
         }
         Ok(acc)
     }
 
-    /// Calculates the product all items (works for all types that can be
+    /// Calculates the product of all items (works for all types that can be
     /// multiplied).
     #[func]
     pub fn product(
@@ -681,6 +692,10 @@ impl Array {
     }
 
     /// Split the array at occurrences of the specified value.
+    ///
+    /// ```example
+    /// #(1, 1, 2, 3, 2, 4, 5).split(2)
+    /// ```
     #[func]
     pub fn split(
         &self,
@@ -697,16 +712,24 @@ impl Array {
     #[func]
     pub fn join(
         self,
-        engine: &mut Engine,
-        span: Span,
         /// A value to insert between each item of the array.
         #[default]
         separator: Option<Value>,
         /// An alternative separator between the last two items.
         #[named]
         last: Option<Value>,
+        /// What to return if the array is empty.
+        #[named]
+        default: Option<Value>,
     ) -> StrResult<Value> {
         let len = self.0.len();
+
+        if let Some(result) = default
+            && len == 0
+        {
+            return Ok(result);
+        }
+
         let separator = separator.unwrap_or(Value::None);
 
         let mut last = last;
@@ -714,18 +737,13 @@ impl Array {
         for (i, value) in self.into_iter().enumerate() {
             if i > 0 {
                 if i + 1 == len && last.is_some() {
-                    result = ops::join(
-                        result,
-                        last.take().unwrap(),
-                        &mut (&mut *engine, span),
-                    )?;
+                    result = ops::join(result, last.take().unwrap())?;
                 } else {
-                    result =
-                        ops::join(result, separator.clone(), &mut (&mut *engine, span))?;
+                    result = ops::join(result, separator.clone())?;
                 }
             }
 
-            result = ops::join(result, value, &mut (&mut *engine, span))?;
+            result = ops::join(result, value)?;
         }
 
         Ok(result)
@@ -733,6 +751,10 @@ impl Array {
 
     /// Returns an array with a copy of the separator value placed between
     /// adjacent elements.
+    ///
+    /// ```example
+    /// #("A", "B", "C").intersperse("-")
+    /// ```
     #[func]
     pub fn intersperse(
         self,
@@ -814,7 +836,7 @@ impl Array {
     /// function. The sorting algorithm used is stable.
     ///
     /// Returns an error if two values could not be compared or if the key
-    /// function (if given) yields an error.
+    /// or comparison function (if given) yields an error.
     ///
     /// To sort according to multiple criteria at once, e.g. in case of equality
     /// between some criteria, the key function can return an array. The results
@@ -834,37 +856,140 @@ impl Array {
         engine: &mut Engine,
         context: Tracked<Context>,
         span: Span,
-        /// If given, applies this function to the elements in the array to
+        /// If given, applies this function to each element in the array to
         /// determine the keys to sort by.
         #[named]
         key: Option<Func>,
+        /// If given, uses this function to compare every two elements in the
+        /// array.
+        ///
+        /// The function will receive two elements in the array for comparison,
+        /// and should return a boolean indicating their order: `{true}`
+        /// indicates that the elements are in order, while `{false}` indicates
+        /// that they should be swapped. To keep the sort stable, if the two
+        /// elements are equal, the function should return `{true}`.
+        ///
+        /// If this function does not order the elements properly (e.g., by
+        /// returning `{false}` for both `{(x, y)}` and `{(y, x)}`, or for
+        /// `{(x, x)}`), the resulting array will be in unspecified order.
+        ///
+        /// When used together with `key`, `by` will be passed the keys instead
+        /// of the elements.
+        ///
+        /// ```example
+        /// #(
+        ///   "sorted",
+        ///   "by",
+        ///   "decreasing",
+        ///   "length",
+        /// ).sorted(
+        ///   key: s => s.len(),
+        ///   by: (l, r) => l >= r,
+        /// )
+        /// ```
+        #[named]
+        by: Option<Func>,
     ) -> SourceResult<Array> {
-        let mut result = Ok(());
-        let mut vec = self.0;
-        let mut key_of = |x: Value| match &key {
-            // NOTE: We are relying on `comemo`'s memoization of function
-            // evaluation to not excessively reevaluate the `key`.
-            Some(f) => f.call(engine, context, [x]),
-            None => Ok(x),
-        };
-        vec.make_mut().sort_by(|a, b| {
-            // Until we get `try` blocks :)
-            match (key_of(a.clone()), key_of(b.clone())) {
-                (Ok(a), Ok(b)) => ops::compare(&a, &b).unwrap_or_else(|err| {
-                    if result.is_ok() {
-                        result = Err(err).at(span);
+        match by {
+            Some(by) => {
+                let mut are_in_order = |mut x, mut y| {
+                    if let Some(f) = &key {
+                        // We rely on `comemo`'s memoization of function
+                        // evaluation to not excessively reevaluate the key.
+                        x = f.call(engine, context, [x])?;
+                        y = f.call(engine, context, [y])?;
                     }
-                    Ordering::Equal
-                }),
-                (Err(e), _) | (_, Err(e)) => {
-                    if result.is_ok() {
-                        result = Err(e);
+                    match by.call(engine, context, [x, y])? {
+                        Value::Bool(b) => Ok(b),
+                        x => {
+                            bail!(
+                                span,
+                                "expected boolean from `by` function, got {}",
+                                x.ty(),
+                            )
+                        }
                     }
-                    Ordering::Equal
-                }
+                };
+                // If a comparison function is provided, we use `glidesort`
+                // instead of the standard library sorting algorithm to prevent
+                // panics in case the comparison function does not define a
+                // valid order (see https://github.com/typst/typst/pull/5627).
+                let mut result = Ok(());
+                let mut vec = self.0.into_iter().enumerate().collect::<Vec<_>>();
+                glidesort::sort_by(&mut vec, |(i, x), (j, y)| {
+                    // Because we use booleans for the comparison function, in
+                    // order to keep the sort stable, we need to compare in the
+                    // right order.
+                    if i < j {
+                        // If `x` and `y` appear in this order in the original
+                        // array, then we should change their order (i.e.,
+                        // return `Ordering::Greater`) iff `y` is strictly less
+                        // than `x` (i.e., `compare(x, y)` returns `false`).
+                        // Otherwise, we should keep them in the same order
+                        // (i.e., return `Ordering::Less`).
+                        match are_in_order(x.clone(), y.clone()) {
+                            Ok(false) => Ordering::Greater,
+                            Ok(true) => Ordering::Less,
+                            Err(err) => {
+                                if result.is_ok() {
+                                    result = Err(err);
+                                }
+                                Ordering::Equal
+                            }
+                        }
+                    } else {
+                        // If `x` and `y` appear in the opposite order in the
+                        // original array, then we should change their order
+                        // (i.e., return `Ordering::Less`) iff `x` is strictly
+                        // less than `y` (i.e., `compare(y, x)` returns
+                        // `false`). Otherwise, we should keep them in the same
+                        // order (i.e., return `Ordering::Less`).
+                        match are_in_order(y.clone(), x.clone()) {
+                            Ok(false) => Ordering::Less,
+                            Ok(true) => Ordering::Greater,
+                            Err(err) => {
+                                if result.is_ok() {
+                                    result = Err(err);
+                                }
+                                Ordering::Equal
+                            }
+                        }
+                    }
+                });
+                result.map(|()| vec.into_iter().map(|(_, x)| x).collect())
             }
-        });
-        result.map(|_| vec.into())
+
+            None => {
+                let mut key_of = |x: Value| match &key {
+                    // We rely on `comemo`'s memoization of function evaluation
+                    // to not excessively reevaluate the key.
+                    Some(f) => f.call(engine, context, [x]),
+                    None => Ok(x),
+                };
+                // If no comparison function is provided, we know the order is
+                // valid, so we can use the standard library sort and prevent an
+                // extra allocation.
+                let mut result = Ok(());
+                let mut vec = self.0;
+                vec.make_mut().sort_by(|a, b| {
+                    match (key_of(a.clone()), key_of(b.clone())) {
+                        (Ok(a), Ok(b)) => ops::compare(&a, &b).unwrap_or_else(|err| {
+                            if result.is_ok() {
+                                result = Err(err).at(span);
+                            }
+                            Ordering::Equal
+                        }),
+                        (Err(e), _) | (_, Err(e)) => {
+                            if result.is_ok() {
+                                result = Err(e);
+                            }
+                            Ordering::Equal
+                        }
+                    }
+                });
+                result.map(|()| vec.into())
+            }
+        }
     }
 
     /// Deduplicates all items in the array.
@@ -873,21 +998,24 @@ impl Array {
     /// element of each duplicate is kept.
     ///
     /// ```example
-    /// #(1, 1, 2, 3, 1).dedup()
+    /// #(3, 3, 1, 2, 3).dedup()
     /// ```
     #[func(title = "Deduplicate")]
     pub fn dedup(
         self,
         engine: &mut Engine,
         context: Tracked<Context>,
-        span: Span,
-        /// If given, applies this function to the elements in the array to
+        /// If given, applies this function to each element in the array to
         /// determine the keys to deduplicate by.
+        ///
+        /// ```example
+        /// #("apple", "banana", " apple ").dedup(key: s => s.trim())
+        /// ```
         #[named]
         key: Option<Func>,
     ) -> SourceResult<Array> {
         let mut out = EcoVec::with_capacity(self.0.len());
-        let key_of = |engine: &mut Engine, x: Value| match &key {
+        let mut key_of = |x: Value| match &key {
             // NOTE: We are relying on `comemo`'s memoization of function
             // evaluation to not excessively reevaluate the `key`.
             Some(f) => f.call(engine, context, [x]),
@@ -898,18 +1026,14 @@ impl Array {
         // 1. We would like to preserve the order of the elements.
         // 2. We cannot hash arbitrary `Value`.
         'outer: for value in self {
-            let key = key_of(&mut *engine, value.clone())?;
+            let key = key_of(value.clone())?;
             if out.is_empty() {
                 out.push(value);
                 continue;
             }
 
             for second in out.iter() {
-                if ops::equal(
-                    &key,
-                    &key_of(&mut *engine, second.clone())?,
-                    &mut (&mut *engine, span),
-                ) {
+                if ops::equal(&key, &key_of(second.clone())?) {
                     continue 'outer;
                 }
             }
@@ -964,6 +1088,11 @@ impl Array {
     /// For arrays with at least one element, this is the same as [`array.fold`]
     /// with the first element of the array as the initial accumulator value,
     /// folding every subsequent element into it.
+    ///
+    /// ```example
+    /// #let array = (2, 1, 4, 3)
+    /// #array.reduce((acc, x) => calc.max(acc, x))
+    /// ```
     #[func]
     pub fn reduce(
         self,
