@@ -8,6 +8,7 @@ mod bitset;
 mod deferred;
 mod duration;
 mod hash;
+mod listset;
 mod pico;
 mod round;
 mod scalar;
@@ -15,7 +16,8 @@ mod scalar;
 pub use self::bitset::{BitSet, SmallBitSet};
 pub use self::deferred::Deferred;
 pub use self::duration::format_duration;
-pub use self::hash::{LazyHash, ManuallyHash};
+pub use self::hash::{HashLock, LazyHash, ManuallyHash};
+pub use self::listset::ListSet;
 pub use self::pico::{PicoStr, ResolvedPicoStr};
 pub use self::round::{round_int_with_precision, round_with_precision};
 pub use self::scalar::Scalar;
@@ -23,11 +25,11 @@ pub use self::scalar::Scalar;
 #[doc(hidden)]
 pub use once_cell;
 
-use std::fmt::{Debug, Formatter};
+use std::fmt::{Debug, Display, Formatter};
 use std::hash::Hash;
 use std::iter::{Chain, Flatten, Rev};
-use std::num::NonZeroUsize;
-use std::ops::{Add, Deref, Div, Mul, Neg, Sub};
+use std::num::{NonZeroU32, NonZeroUsize};
+use std::ops::{Add, Deref, DerefMut, Div, Mul, Neg, Sub};
 use std::sync::Arc;
 
 use siphasher::sip128::{Hasher128, SipHasher13};
@@ -52,6 +54,25 @@ where
     Wrapper(f)
 }
 
+/// Turn a closure into a struct implementing [`Display`].
+pub fn display<F>(f: F) -> impl Display
+where
+    F: Fn(&mut Formatter) -> std::fmt::Result,
+{
+    struct Wrapper<F>(F);
+
+    impl<F> Display for Wrapper<F>
+    where
+        F: Fn(&mut Formatter) -> std::fmt::Result,
+    {
+        fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
+            self.0(f)
+        }
+    }
+
+    Wrapper(f)
+}
+
 /// Calculate a 128-bit siphash of a value.
 pub fn hash128<T: Hash + ?Sized>(value: &T) -> u128 {
     let mut state = SipHasher13::new();
@@ -66,10 +87,11 @@ pub trait NonZeroExt {
 }
 
 impl NonZeroExt for NonZeroUsize {
-    const ONE: Self = match Self::new(1) {
-        Some(v) => v,
-        None => unreachable!(),
-    };
+    const ONE: Self = Self::new(1).unwrap();
+}
+
+impl NonZeroExt for NonZeroU32 {
+    const ONE: Self = Self::new(1).unwrap();
 }
 
 /// Extra methods for [`Arc`].
@@ -360,6 +382,58 @@ pub fn default_math_class(c: char) -> Option<MathClass> {
         // https://github.com/typst/typst/pull/5714
         '\u{22A5}' => Some(MathClass::Normal),
 
+        // Used as a binary connector in linear logic, where it is referred to
+        // as "par".
+        // https://github.com/typst/typst/issues/5764
+        '⅋' => Some(MathClass::Binary),
+
+        // Those overrides should become the default in the next revision of
+        // MathClass.txt.
+        // https://github.com/typst/typst/issues/5764#issuecomment-2632435247
+        '⎰' | '⟅' => Some(MathClass::Opening),
+        '⎱' | '⟆' => Some(MathClass::Closing),
+
+        // Both ∨ and ⟑ are classified as Binary.
+        // https://github.com/typst/typst/issues/5764
+        '⟇' => Some(MathClass::Binary),
+
+        // Arabic comma.
+        // https://github.com/latex3/unicode-math/pull/633#issuecomment-2028936135
+        '،' => Some(MathClass::Punctuation),
+
         c => unicode_math_class::class(c),
     }
 }
+
+/// Automatically calls a deferred function when the returned handle is dropped.
+pub fn defer<T, F: FnOnce(&mut T)>(
+    thing: &mut T,
+    deferred: F,
+) -> impl DerefMut<Target = T> {
+    pub struct DeferHandle<'a, T, F: FnOnce(&mut T)> {
+        thing: &'a mut T,
+        deferred: Option<F>,
+    }
+
+    impl<'a, T, F: FnOnce(&mut T)> Drop for DeferHandle<'a, T, F> {
+        fn drop(&mut self) {
+            std::mem::take(&mut self.deferred).expect("deferred function")(self.thing);
+        }
+    }
+
+    impl<T, F: FnOnce(&mut T)> std::ops::Deref for DeferHandle<'_, T, F> {
+        type Target = T;
+
+        fn deref(&self) -> &Self::Target {
+            self.thing
+        }
+    }
+
+    impl<T, F: FnOnce(&mut T)> std::ops::DerefMut for DeferHandle<'_, T, F> {
+        fn deref_mut(&mut self) -> &mut Self::Target {
+            self.thing
+        }
+    }
+
+    DeferHandle { thing, deferred: Some(deferred) }
+}
